#ifndef _BITMAP_H
#define _BITMAP_H


#if defined(__SIZEOF_LONG__) && (__SIZEOF_LONG__ == 8)
#define BYTES_PER_LONG	8
#define BITMAP_TOP_BIT	0x80000000
#else
#define BYTES_PER_LONG	4
#define BITMAP_TOP_BIT	0x8000
#endif

#define BITS_PER_LONG	(BYTES_PER_LONG<<3)
#define BITMAP_SIZE(bits) (((bits) + (BITS_PER_LONG)) / (BITS_PER_LONG))


#define BITMAP(name, size) 			\
	unsigned long name[BITMAP_SIZE(size)];


/**
 * bitmap_create
 * \brief	initialize a new bitmap with a range from 0 -> (size - 1)
 * \param	bm	pointer to a bitmap struct
 * \param[in]	size	size
 */
int bitmap_create( unsigned long **bm, const ssize_t size );

/**
 * bitmap_is_set
 * \brief	Check whether a bit in the map is set
 * \param	bm	bitmap
 * \param[in]	pos	bit to check
 * \return	1 if the bit is set, 0 if it's not, -1 on error
 */
int bitmap_is_set( unsigned long *bm, const int nbits, const int pos );

/**
 * bitmap_set
 * \brief	Set a bit in the map
 * \param	bm	bitmap
 * \param[in]	pos	bit to set
 * \return	0 on success, -1 on error
 */
int bitmap_set( unsigned long *bm, const int nbits, const int pos );

/**
 * bitmap_clear
 * \brief	Clear a bit in the map
 * \param	bm	bitmap
 * \param[in]	pos	bit to clear
 * \return	0 on success, -1 on error
 */
int bitmap_clear( unsigned long *bm, const int nbits, const int pos );

/**
 * bitmap_clear_all
 * \brief	Clear a map completely
 * \param	bm	bitmap to clear
 */
void bitmap_clear_all( unsigned long *bm, const int nbits );

/**
 * bitmap_is_empty
 * \brief	Check if map has no entries
 * \param[in]	bm	bitmap
 * \return	0 if map has entries, 1 otherwise
 */
int bitmap_is_empty( const unsigned long *bm, const int nbits );


/**
 * bitmap_first_bit
 * \brief	Return nr of first set bit in map
 * \param[in]	bm	bitmap
 * \return	0 if none found, >0 nr of first bit
 */
int bitmap_first_bit( const unsigned long *bm, const int nbits );


#endif
