#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

#include <pthread.h>

#include "bitmap.h"
#include "trie.h"

#define tv_diff(a, b) ({ \
		long long int __tmpA, __tmpB; \
		__tmpA = (a.tv_sec * 1000000) + a.tv_usec; \
		__tmpB = (b.tv_sec * 1000000) + b.tv_usec; \
		__tmpA - __tmpB; \
	})

#define ts_diff(a, b) ({ \
		long long int __tmpA, __tmpB; \
		__tmpA = (a.tv_sec * 1000000) + (a.tv_nsec / 1000); \
		__tmpB = (b.tv_sec * 1000000) + (b.tv_nsec / 1000); \
		__tmpA - __tmpB; \
	})

#define NUM_ITER	1000000
#define NUM_ROUTES	100000
#define NUM_THREADS	4

void dump_cb(const char *key, const int keylen, const void *val, const void *priv)
{
	char buf[TRIE_MAX_DEPTH+1];
	int x;

	for( x=0; x<keylen; x++ ) {
		buf[x] = 48 + key[x];
	}
	buf[keylen] = '\0';

	fprintf( (FILE *)priv, "%-30s\t\t%d\n", buf, (val) ? *(int *)val : 0 );
}

void traverse_cb(const struct trie_node *n, const char *key, const int keylen, const void *value, void *priv)
{
#if 0
	char buf[TRIE_MAX_DEPTH+1];
	int x;

	for( x=0; x<keylen; x++ ) {
		buf[x] = 48 + key[x];
	}
	buf[keylen] = '\0';

	printf( "key %-30s len %d (%p)\n", buf, keylen, n );
#endif
	(*(int *)priv)++;
}

#define lpm_test3_quiet(_key, _len) ({						\
	char *_x = _key;							\
	trie_get_partial( &lpm_trie, _x, _len );				\
	num_routes++;								\
	sum_length+= _len;							\
})

#define lpm_test4_quiet(_trie, _key, _len) ({					\
	char *_x = _key;							\
	trie_get_partial( _trie, _x, _len );					\
	num_routes++;								\
	sum_length+= _len;							\
})


struct thread_arg {
	struct trie_node	*trie;

	pthread_barrier_t	*wait;

	/* stats */
	float	tdiff;
	int	sum_length, num_routes;
};

void *search_thread(void *arg)
{
	int x=0, y=0, sum_length=0, num_routes=0, len=0;
	struct thread_arg *t = arg;
	struct timespec ts_before, ts_after;
	float tdiff = 0.0f;
	char buf[TRIE_MAX_DEPTH];

	pthread_barrier_wait( t->wait );

	clock_gettime( CLOCK_REALTIME, &ts_before );
	for(x=0; x<NUM_ITER; x++) {
		len = 20 + random() % 10;

		for( y=0; y<len; y++ )
			buf[y] = random() % 10;

		lpm_test4_quiet( t->trie, buf, len );
	}

	clock_gettime( CLOCK_REALTIME, &ts_after );

	tdiff = (float)ts_diff( ts_after, ts_before );

	t->num_routes = num_routes;
	t->tdiff = tdiff;
	t->sum_length = sum_length;

	return NULL;
}

int main( void )
{
	int x=0, y=0, len=0;
	int sum_length=0, num_routes=0;

	struct timespec ts_before, ts_after;
	char buf[TRIE_MAX_DEPTH];
	int  hist[TRIE_MAX_DEPTH];
	float tdiff = 0.0f;

	int val1 = 12345;

	printf( "\n\nTests:\n" );

#if 0
	/*
	 * Test 13: VoIP related longest prefix matching benchmark + tests
	 */
	printf( "\n*** Test 13: LMP Benchmark, single threaded\n" );
	{
		struct trie_node lpm_trie;
		FILE *fp = NULL;
		int trav_counter = 0;

		/*
		 * 13.4: speed
		 */
		printf( "\n  * Part 4: LPM Route lookup tests (no output on successfull match)\n" );
		long len = 0;

		memset( hist, 0, TRIE_MAX_DEPTH * sizeof(int) );

		trie_init( &lpm_trie );

		// add routes
		for( x=0; x<NUM_ROUTES; ) {
			len = random() % TRIE_MAX_DEPTH;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			if( trie_add( &lpm_trie, buf, len, &val1 ) == 0 ) {
				hist[len] += 1;
				x++;
			}
		}
		printf( "   %d prefixes added\n", x );

		printf( "\n   --- Histogram ---\n" );
		for( x = 0; x<TRIE_MAX_DEPTH; x++ )
			printf( "\t%03d: %d\n", x, hist[x] );

			
		clock_gettime( CLOCK_REALTIME, &ts_before );
		for(x=0; x<NUM_ITER; x++) {
			len = 20 + random() % 10;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			lpm_test3_quiet( buf, len );
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes));
		printf( "   Time for lookups           : %-5.3f usecs\n", tdiff );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );



		printf( "\n\n ------ 1 second benchmark -------\n" );
		num_routes = 0;
		sum_length = 0;
		tdiff = 0.0f;

		clock_gettime( CLOCK_REALTIME, &ts_before );
		for( x=0; tdiff < 1000000.0f; x++ ) {
			len = 20 + random() % 10;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			lpm_test3_quiet( buf, len );

			if( x % 1000 == 0 ) {
				clock_gettime( CLOCK_REALTIME, &ts_after );
				tdiff = (float)ts_diff( ts_after, ts_before );
			}
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes ); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes) );
		printf( "   Time for lookups           : %-3.6f secs\n", tdiff / 1000000.0f );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );



		/*
		 * 14: speed of traversing all end nodes
		 */
		printf( "\n*** Test 14: Time for traversing a %d key trie\n", NUM_ROUTES );

		/*
		 * write stuff to disk
		 */
		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_traverse( &lpm_trie, &traverse_cb, &trav_counter, TRIE_TRAVERSE_TERM );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "   Nodes traversed                : %d\n", trav_counter );
		printf( "   Time for traversing            : %-5.3f usecs\n", tdiff );
		printf( "   Avg. time per node             : %-5.3f usecs\n", (tdiff/NUM_ROUTES) );




		/*
		 * 15: speed of destroying
		 */
		printf( "\n*** Test 15: Time for destroying a (max.) %d key trie\n", NUM_ROUTES );

		/*
		 * write stuff to disk
		 */
		fp = fopen( "trie.dmp", "w+" );
		if( !fp ) {
			perror( "opening trie.dmp" );
			return 1;
		}

		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_dump_destroy( &lpm_trie, &dump_cb, fp );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		fclose( fp );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "   Time for destruction (w/ dump) : %-5.3f usecs\n", tdiff );
		printf( "   Avg. time per node             : %-5.3f usecs\n", (tdiff/NUM_ROUTES) );

		/* sanity check */
		for( x=0; x<TRIE_NUM_SLOTS; x++ ) {
			if( lpm_trie.children[x] )
				break;
		}
		if( x < TRIE_NUM_SLOTS )
			printf( "!!! Root node still has children\n" );
	}
#endif

	/*
	 * Test 13b: VoIP related longest prefix matching benchmark + tests
	 */
	printf( "\n*** Test 13b: LMP Benchmark, single threaded\n" );
	{
		struct trie_node lpm_trie;
		FILE *fp = NULL;
		int trav_counter = 0;

		/*
		 * 13.4: speed
		 */
		printf( "\n  * Part 4: LPM Route lookup tests (no output on successfull match)\n" );
		long len = 0;

		memset( hist, 0, TRIE_MAX_DEPTH * sizeof(int) );

		trie_init( &lpm_trie );

		// add routes
		for( x=0; x<NUM_ROUTES; ) {
			len = random() % TRIE_MAX_DEPTH;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			if( trie_add( &lpm_trie, buf, len, &val1 ) == 0 ) {
				hist[len] += 1;
				x++;
			}
		}
		printf( "   %d prefixes added\n", x );

		printf( "\n   --- Histogram ---\n" );
		for( x = 0; x<TRIE_MAX_DEPTH; x++ )
			printf( "\t%03d: %d\n", x, hist[x] );


		// del routes
		for( x = 0; x < (NUM_ROUTES / 10); ) {
			len = ( random() + random() ) % TRIE_MAX_DEPTH;

			for( y = 0; y < len; y++ )
				buf[y] = random() % 10;

			if( !trie_has_key( &lpm_trie, buf, len) )
				continue;

			if( trie_del( &lpm_trie, buf, len ) == 0 ) {
				hist[len] -= 1;
				x++;
				printf( "%d routes removed, %d left\r", x, (NUM_ROUTES / 10) - x );
			}
		}
		printf( "   %d prefixes removed\n", x );

		printf( "\n   --- Histogram ---\n" );
		for( x = 0; x<TRIE_MAX_DEPTH; x++ )
			printf( "\t%03d: %d\n", x, hist[x] );
			
		clock_gettime( CLOCK_REALTIME, &ts_before );
		for(x=0; x<NUM_ITER; x++) {
			len = 20 + random() % 10;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			lpm_test3_quiet( buf, len );
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes));
		printf( "   Time for lookups           : %-5.3f usecs\n", tdiff );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );



		printf( "\n\n ------ 1 second benchmark -------\n" );
		num_routes = 0;
		sum_length = 0;
		tdiff = 0.0f;

		clock_gettime( CLOCK_REALTIME, &ts_before );
		for( x=0; tdiff < 1000000.0f; x++ ) {
			len = 20 + random() % 10;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			lpm_test3_quiet( buf, len );

			if( x % 1000 == 0 ) {
				clock_gettime( CLOCK_REALTIME, &ts_after );
				tdiff = (float)ts_diff( ts_after, ts_before );
			}
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes ); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes) );
		printf( "   Time for lookups           : %-3.6f secs\n", tdiff / 1000000.0f );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );


		/*
		 * 14: speed of traversing all end nodes
		 */
		printf( "\n*** Test 14: Time for traversing a %d key trie\n", NUM_ROUTES );

		/*
		 * write stuff to disk
		 */
		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_traverse( &lpm_trie, &traverse_cb, &trav_counter, TRIE_TRAVERSE_TERM );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "   Nodes traversed                : %d\n", trav_counter );
		printf( "   Time for traversing            : %-5.3f usecs\n", tdiff );
		printf( "   Avg. time per node             : %-5.3f usecs\n", (tdiff/NUM_ROUTES) );


		/*
		 * 15: speed of destroying
		 */
		printf( "\n*** Test 15: Time for destroying a (max.) %d key trie\n", NUM_ROUTES );

		/*
		 * write stuff to disk
		 */
		fp = fopen( "trie2.dmp", "w+" );
		if( !fp ) {
			perror( "opening trie.dmp" );
			return 1;
		}

		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_dump_destroy( &lpm_trie, &dump_cb, fp );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		fclose( fp );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "   Time for destruction (w/ dump) : %-5.3f usecs\n", tdiff );
		printf( "   Avg. time per node             : %-5.3f usecs\n", (tdiff/NUM_ROUTES) );

		/* sanity check */
		for( x = 0;  x < TRIE_NUM_SLOTS; x++ ) {
			if( lpm_trie.children[x] )
				break;
		}
		if( x < TRIE_NUM_SLOTS )
			printf( "!!! Root node still has children\n" );
	}



	/*
	 * Same, multithreaded
	 */
	printf( "\n*** Test 16: LMP Benchmark, multithreaded, %d threads\n", NUM_THREADS );
	{
		struct trie_node lpm_trie;
		pthread_t thread[NUM_THREADS];
		struct thread_arg args[NUM_THREADS];
		pthread_barrier_t barrier;

		memset( hist, 0, TRIE_MAX_DEPTH * sizeof(int) );

		trie_init( &lpm_trie );

		// add routes
		for( x=0; x<NUM_ROUTES; ) {
			len = random() % TRIE_MAX_DEPTH;

			for( y=0; y<len; y++ )
				buf[y] = random() % 10;

			if( trie_add( &lpm_trie, buf, len, &val1 ) == 0 ) {
				hist[len] += 1;
				x++;
			}
		}
		printf( "   %d prefixes added\n", x );

		printf( "\n   --- Histogram ---\n" );
		for( x = 0; x<TRIE_MAX_DEPTH; x++ )
			printf( "\t%03d: %d\n", x, hist[x] );
		

		printf( "   Starting %d search threads\n", NUM_THREADS );

		pthread_barrier_init( &barrier, NULL, NUM_THREADS );

		for( x=0; x<NUM_THREADS; x++ ) {
			args[x].trie = &lpm_trie;
			args[x].wait = &barrier;

			pthread_create( &thread[x], NULL, &search_thread, &args[x] );
		}

		clock_gettime( CLOCK_REALTIME, &ts_before );
		for( x=0; x<NUM_THREADS; x++ ) {
			pthread_join( thread[x], NULL );
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		for( x=0; x<NUM_THREADS; x++ ) {
			struct thread_arg *t = &args[x];

			/* output some stats */
			printf( "\n   --- Stats Thread %d ---\n", x );
			printf( "   Number of lookups performed: %d keys\n", t->num_routes ); 
			printf( "   Average route key length   : %.2f chars\n", ((float)t->sum_length/(float)t->num_routes) );
			printf( "   Time for lookups           : %-3.6f secs\n", t->tdiff / 1000000.0f );
			printf( "   Average time per lookup    : %-5.3f usecs\n", (t->tdiff/(float)t->num_routes) );
		}

		pthread_barrier_destroy( &barrier );


		printf( "\n   Destroying trie\n" );

		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_destroy( &lpm_trie );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before );

		/* output some stats */
		printf( "   Time for destruction (w/o dump): %-5.3f usecs\n", tdiff );
		printf( "   Avg. time per node             : %-5.3f usecs\n", (tdiff/NUM_ROUTES) );

	}

	return 0;
}

